import time
from pprint import pprint
from pathlib import Path
from jenky import util

config = [
    {
        "repoName": "jenky",
        "directory": "jenky",
        "gitTag": "4711",
        "gitTags": ["..."],
        "processes": [
            {
                "name": "jenky",
                "running": None,
                "cmdPattern": {
                    "index": 1,
                    "pattern": "test_jenky"
                }
            }
        ]
    }
]


def test_fill_process_running():
    util.fill_process_running(config)
    pprint(config)


async def test_fill_tag():
    await util.fill_git_tag(config)
    pprint(config)


def test_fill_tags():
    util.fill_git_tags(config)
    pprint(config)


def test_find_root():
    procs = [
        dict(pid=1, ppid=2, info='p1'),
        dict(pid=2, ppid=3, info='p2')]
    root = util.find_root(procs)
    assert root == procs[1]

    procs = [
        dict(pid=2, ppid=3, info='p2'),
        dict(pid=1, ppid=2, info='p1')
        ]
    root = util.find_root(procs)
    assert root == procs[0]

    procs = [
        dict(pid=1, ppid=2, info='p1'),
        dict(pid=3, ppid=4, info='p2')
    ]
    try:
        root = util.find_root(procs)
        assert False
    except AssertionError:
        pass


def test_git_branches():
    branches = util.git_branches(Path('c:/ws/projects/VirtualPowerStorage'))
    pprint(branches)


def test_git_tags():
    tags = util.git_tags(Path('c:/ws/projects/VirtualPowerStorage'))
    pprint(tags)


def test_run():
    cwd = Path('.').absolute()
    util.run(cwd, ['bash', 'p.sh'])
    time.sleep(1)
    util.dump_processes([cwd])


def test_match_cmd():
    cmd1 = ['bash', 'foo.sh']
    cmd2 = [r'C:\\WINDOWS\\system32\\wsl.exe', '-e', '/bin/bash', 'foo.sh']
    assert util.match_cmd(cmd1, cmd2)


util.git_cmd = 'C:/ws/tools/PortableGit/bin/git.exe'
test_match_cmd()



import json
import logging
import sys
import time
from pathlib import Path
from typing import List

import uvicorn
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from pydantic import BaseModel
from starlette.responses import RedirectResponse, Response

from jenky import util
from jenky.util import Config, Repo, get_tail

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
handler.formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s %(funcName)s - %(message)s')
logger.addHandler(handler)

app = FastAPI()
html_root = Path(__file__).parent / 'html'
app.mount("/static", StaticFiles(directory=html_root), name="mymountname")


@app.get("/")
def home():
    return RedirectResponse(url='/static/index.html')


@app.get("/repos")
def get_repos() -> Config:
    util.running_processes(config.repos)
    return config


@app.get("/repos/{repo_id}")
def get_repo(repo_id: str) -> Repo:
    repo = util.repo_by_id(config.repos, repo_id)
    # util.fill_git_tag(config.repos)
    util.fill_git_tags(repo)
    util.fill_git_branches(repo)
    return repo


class Action(BaseModel):
    action: str


@app.post("/repos/{repo_id}/processes/{process_id}")
def post_process(repo_id: str, process_id: str, action: Action):
    if action.action == 'kill':
        util.kill(config.repos, repo_id, process_id)
    elif action.action == 'restart':
        util.restart(config.repos, repo_id, process_id)
        time.sleep(1)
    else:
        assert False, 'Invalid action ' + action.action

    return dict(repo_id=repo_id, process_id=process_id, action=action.action)


@app.get("/repos/{repo_id}/processes/{process_id}/{std_x}")
def get_process_log(repo_id: str, process_id: str, std_x: str) -> Response:
    repo = util.repo_by_id(config.repos, repo_id)
    path = util.base_url / repo.directory / f'{process_id}.{std_x[3:]}'

    lines = get_tail(path)
    return Response(content=''.join(lines), media_type="text/plain")


class GitAction(BaseModel):
    action: str
    gitTag: str


@app.post("/repos/{repo_id}")
def post_repo(repo_id: str, action: GitAction):
    if action.action == 'pull':
        repo = util.repo_by_id(config.repos, repo_id)
        message = util.git_pull(repo)  # , target_tag=action.gitTag)
    else:
        assert False, 'Invalid action ' + action.action

    return dict(repo_id=repo_id, action=action.action, message=message)


config_file = 'config.json'
host = "127.0.0.1"
port = 8000

# TODO: Use argparse
for arg in sys.argv:
    if arg.startswith('--config='):
        config_file = arg.split('=')[1]
    elif arg.startswith('--port='):
        port = int(arg.split('=')[1])
    elif arg.startswith('--host='):
        host = arg.split('=')[1]

config_file = Path(config_file)
logger.info(f'Reading config from {config_file}')
config = Config.parse_obj(json.loads(config_file.read_text(encoding='utf8')))
util.git_cmd = config.git_cmd
util.base_url = config_file.parent.absolute()

uvicorn.run(app, host=host, port=port)



from collections import defaultdict
from pathlib import Path
from typing import List, Dict

import psutil


def dump_processes(dirs: List[Path]) -> Dict[str, dict]:
    print(f'dump_processes {dirs}')
    processes_by_pid = {}
    username = psutil.Process().username()

    attrs = ['pid', 'ppid', 'name', 'cwd', 'exe', 'username', 'cmdline', 'create_time']

    for proc in psutil.process_iter(attrs=attrs, ad_value=None):
        info = proc.info
        if info.get('username', None) != username:
            continue
        if Path(info.get('cwd', '') or '') not in dirs:
            continue
        processes_by_pid[info['pid']] = info

    # return
    children_by_pid = defaultdict(lambda: [])
    for n in processes_by_pid.values():
        children_by_pid[n['ppid']].append(n)

    def taint_parents(node):
        # print(f'taint_parents {node["pid"]}')
        while node['ppid'] in processes_by_pid:
            node = processes_by_pid[node['ppid']]
            if 'taint' in node:
                break
            node['taint'] = True

    def taint_children(node):
        # print(f'taint_children {node["pid"]}')
        for child in children_by_pid[node['pid']]:
            if 'taint' not in child:
                child['taint'] = True
                taint_children(child)

    for node in processes_by_pid.values():
        if True or 'cwd' in node and node['cwd'] and Path(node['cwd']) in dirs:
            taint_children(node)
            taint_parents(node)
            node['taint'] = True

    def dump(node, indent, verbose):
        # print(f'dump {node["pid"]}')
        pid = node['pid']
        ppid = node['ppid']
        name = node['name']
        username = node['username']
        cwd = node.get('cwd', '')
        paths = [f.path for f in (node.get('open_files', []) or [])]
        cmdline = node.get('cmdline', []) or []
        print(indent + f'{pid} {ppid} {username} {name} {cwd} ' + ', '.join(cmdline) + ' ' + ', '.join(paths))
        # print(indent + f'{pid} {ppid} ' + str(node))
        for child in children_by_pid[pid]:
            if 'taint' in child:
                dump(child, indent + '  ', verbose)

    roots = [node for node in processes_by_pid.values() if node['ppid'] not in processes_by_pid and 'taint' in node]
    for root in roots:
        dump(root, '', True)

    return processes_by_pid



# Note: FastApi does not support asyncio subprocesses, so do not use it!
import logging
import os
from pathlib import Path
from typing import List
import subprocess

import psutil
from pydantic import BaseModel, Field

logger = logging.getLogger()

# git_cmd = 'C:/ws/tools/PortableGit/bin/git.exe'
# git_cmd = 'git'
git_cmd: str
base_url: Path


class Process(BaseModel):
    name: str
    cmd: List[str]
    env: dict
    running: bool
    create_time: float = Field(..., alias='createTime')


class Repo(BaseModel):
    repoName: str
    directory: str
    git_tag: str = Field(..., alias='gitTag')
    git_tags: List[str] = Field(..., alias='gitTags')
    git_branches: List[str] = Field(..., alias='gitBranches')
    git_message: str = Field(..., alias='gitMessage')
    processes: List[Process]


class Config(BaseModel):
    repos: List[Repo]
    git_cmd: str


def running_processes(repos: List[Repo]):
    for repo in repos:
        for proc in repo.processes:
            proc.running = False
            pid_file = base_url / repo.directory / (proc.name + '.pid')
            logger.debug(f'{pid_file}')
            if not pid_file.exists():
                logger.debug(f'Skipping {pid_file}')
                continue

            try:
                pid = int(pid_file.read_text())
            except Exception as e:
                logger.exception('Reading pid file')
                raise e

            try:
                p = psutil.Process(pid)
                proc.running = p.is_running()
            except psutil.NoSuchProcess:
                logger.debug(f'No such proccess {pid_file} {pid}')


def git_tag(git_dir: Path) -> str:
    logger.debug(git_dir)
    proc = subprocess.run(
        [git_cmd, 'describe', '--tags'],
        cwd=git_dir.as_posix(),
        capture_output=True)

    if proc.stderr:
        raise OSError(str(proc.stderr, encoding='ascii'))
    tag = str(proc.stdout, encoding='ascii')
    return tag


def fill_git_tag(repos: List[Repo]):
    for repo in repos:
        try:
            git_dir = base_url / repo.directory
            repo.git_tag = git_tag(git_dir)
        except OSError as e:
            repo.git_tag = None
            repo.git_message = str(e)


git_format = "%(refname:short) %(authorname) %(authordate:raw)"


def git_tags(git_dir: Path) -> List[str]:
    logger.debug(git_dir)
    proc = subprocess.run(
        [git_cmd, 'tag', '--sort', 'version:refname', f"--format={git_format}"],
        cwd=git_dir.as_posix(),
        capture_output=True)

    if proc.stderr:
        raise OSError(str(proc.stderr, encoding='utf8'))
    tags = [line.strip() for line in str(proc.stdout, encoding='utf8').splitlines()]
    return tags


def git_branches(git_dir: Path) -> List[str]:
    logger.debug(git_dir)
    proc = subprocess.run(
        [git_cmd, 'branch', '--sort=-committerdate', f"--format=%(HEAD) {git_format}"],
        cwd=git_dir.as_posix(),
        capture_output=True)

    if proc.stderr:
        raise OSError(str(proc.stderr, encoding='utf8'))
    branches = [line.strip() for line in str(proc.stdout, encoding='utf8').splitlines()]
    return branches


def fill_git_tags(repo: Repo):
    try:
        git_dir = base_url / repo.directory
        repo.git_tags = git_tags(git_dir)
    except OSError as e:
        repo.git_tags = []
        repo.git_message = str(e)


def fill_git_branches(repo: Repo):
    try:
        git_dir = base_url / repo.directory
        repo.git_branches = git_branches(git_dir)
    except OSError as e:
        repo.git_branches = []
        repo.git_message = str(e)


def git_pull(repo: Repo) -> str:
    """
    git pull
    """
    git_dir = base_url / repo.directory
    cmd = [git_cmd, 'pull']
    logger.debug(f'{git_dir} {cmd}')
    proc = subprocess.run(
        cmd,
        cwd=git_dir.as_posix(),
        capture_output=True)

    message = str(proc.stderr, encoding='ascii').rstrip()
    message += str(proc.stdout, encoding='ascii').rstrip()

    # if repo.git_tag != target_tag:
    #     proc = subprocess.run(
    #         [git_cmd, 'checkout', target_tag],
    #         cwd=git_dir.as_posix(),
    #         capture_output=True)
    #     message += '\n' + str(proc.stdout, encoding='ascii').rstrip()

    return message


def run(name: str, cwd: Path, cmd: List[str], env: dict):
    logger.debug(f'Running: {" ".join(cmd)}')
    my_env = os.environ
    my_env.update(env)
    kwargs = {}
    kwargs.update(start_new_session=True)
    pid = subprocess.Popen(
        cmd,
        stdout=open((cwd / f'{name}.out').as_posix(), 'w'),
        stderr=subprocess.STDOUT,
        # stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
        cwd=cwd.as_posix(),
        env=my_env,
        **kwargs).pid

    pid_file = base_url / cwd / (name + '.pid')
    pid_file.write_text(str(pid))


def kill(repos: List[Repo], repo_id: str, process_id: str):
    repo = repo_by_id(repos, repo_id)
    procs = [proc for proc in repo.processes if proc.name == process_id]
    if not procs:
        raise ValueError(repo_id)
    proc = procs[0]
    pid_file = base_url / repo.directory / (proc.name + '.pid')
    pid = int(pid_file.read_text())

    try:
        proc = psutil.Process(pid)
    except psutil.NoSuchProcess:
        logger.warning(f'No such process with pid {pid}')
        return

    proc.terminate()
    gone, alive = psutil.wait_procs([proc], timeout=3, callback=None)
    for p in alive:
        p.kill()


def repo_by_id(repos: List[Repo], repo_id: str) -> Repo:
    repos = [repo for repo in repos if repo.repoName == repo_id]
    if not repos:
        raise ValueError(repo_id)
    return repos[0]


def restart(repos: List[Repo], repo_id: str, process_id: str):
    repo = repo_by_id(repos, repo_id)
    procs = [proc for proc in repo.processes if proc.name == process_id]
    if not procs:
        raise ValueError(repo_id)
    proc = procs[0]
    run(proc.name, base_url / repo.directory, proc.cmd, proc.env)


# def find_root(procs: List[dict]):
#     parent = None
#     for proc in procs:
#         parents = [p for p in procs if p['pid'] == proc['ppid']]
#
#         if not parents:
#             assert parent is None
#             parent = proc
#         else:
#             assert len(parents) == 1
#
#     return parent

def get_tail(path: Path) -> List[str]:
    logger.debug(path)
    with open(path.as_posix(), "rb") as f:
        try:
            f.seek(-10*1024, os.SEEK_END)
            byte_lines = f.readlines()
            if len(byte_lines):
                byte_lines = byte_lines[1:]
            else:
                # So we are in the middle of a line and could hit a composed unicode character.
                # But we just ignore that...
                pass
        except:
            # file size too short
            f.seek(0)
            byte_lines = f.readlines()
    lines = [str(byte_line, encoding='utf8') for byte_line in byte_lines]
    return lines



"""
"""
from setuptools import setup
package_data = {'': ['*'], 'jenky': ['html/*']}

dist = setup(
    name='jenky',
    version='0.0.1',
    author="Wolfgang Kühn",
    description=("A build and deploy server for Python developers"),
    packages=['jenky'],
    # include_package_data=True,
    install_requires=['aiofiles', 'fastapi', 'psutil', 'uvicorn'],
    extras_require={},
    package_data=package_data
)
